import os
import posixpath
import logging
try:
    from urllib.parse import unquote, urlparse
except ImportError:
    from urllib import unquote
    from urlparse import urlparse
try:
    from http.server import HTTPServer, SimpleHTTPRequestHandler
except ImportError:
    from BaseHTTPServer import HTTPServer
    from SimpleHTTPServer import SimpleHTTPRequestHandler

from avocado.utils.astring import to_text


class HTTPRequestHandler(SimpleHTTPRequestHandler):

    def do_GET(self):
        """
        Serve a GET request.
        """
        rg = self.parse_header_byte_range()
        if rg:
            f = self.send_head_range(rg[0], rg[1])
            if f:
                self.copyfile_range(f, self.wfile, rg[0], rg[1])
                f.close()
        else:
            f = self.send_head()
            if f:
                self.copyfile(f, self.wfile)
                f.close()

    def parse_header_byte_range(self):
        range_param = 'Range'
        range_discard = 'bytes='
        if range_param in self.headers:
            rg = self.headers.get(range_param)
            if rg.startswith(range_discard):
                rg = rg[len(range_discard):]
                begin, end = rg.split('-')
                return (int(begin), int(end))
        return None

    def copyfile_range(self, source_file, output_file, range_begin, range_end):
        """
        Copies a range of a file to destination.
        """
        range_size = range_end - range_begin + 1
        source_file.seek(range_begin)
        buf = source_file.read(range_size)
        output_file.write(buf)

    def send_head_range(self, range_begin, range_end):
        path = self.translate_path(self.path)
        f = None
        if os.path.isdir(path):
            for index in "index.html", "index.htm":
                index = os.path.join(path, index)
                if os.path.exists(index):
                    path = index
                    break
            else:
                return self.list_directory(path)
        ctype = self.guess_type(path)
        try:
            # Always read in binary mode. Opening files in text mode may cause
            # newline translations, making the actual size of the content
            # transmitted *less* than the content-length!
            f = open(path, 'rb')
        except IOError:
            self.send_error(404, "File not found")
            return None
        self.send_response(206, "Partial Content")
        file_size = str(os.fstat(f.fileno())[6])
        range_size = str(range_end - range_begin + 1)
        self.send_header("Accept-Ranges", "bytes")
        self.send_header("Content-Length", range_size)
        self.send_header("Content-Range", "bytes %s-%s/%s" % (range_begin,
                                                              range_end,
                                                              file_size))
        self.send_header("Content-type", ctype)
        self.end_headers()
        return f

    def translate_path(self, path):
        """
        Translate a /-separated PATH to the local filename syntax.

        Components that mean special things to the local file system
        (e.g. drive or directory names) are ignored.  (XXX They should
        probably be diagnosed.)

        """
        # abandon query parameters
        path = urlparse(to_text(path))[2]
        path = posixpath.normpath(unquote(path))
        words = path.split('/')
        words = list(filter(None, words))
        path = self.server.cwd
        for word in words:
            _, word = os.path.splitdrive(word)
            _, word = os.path.split(word)
            if word in (os.curdir, os.pardir):
                continue
            path = os.path.join(path, word)
        return path

    def address_string(self):
        '''
        This HTTP server does not care about name resolution for the requests

        The first reason is that most of the times our clients are going to be
        virtual machines without a proper name resolution setup. Also, by not
        resolving names, we should be a bit faster and be resilient about
        misconfigured or resilient name servers.
        '''
        return self.client_address[0]

    def log_message(self, fmt, *args):
        logging.debug("builtin http server handling request from %s: %s" %
                      (self.address_string(), fmt % args))


def http_server(port=8000, cwd=None, terminate_callable=None):
    http = HTTPServer(('', port), HTTPRequestHandler)
    http.timeout = 1

    if cwd is None:
        cwd = os.getcwd()
    http.cwd = cwd

    while True:
        if terminate_callable is not None:
            terminate = terminate_callable()
        else:
            terminate = False

        if terminate:
            break

        http.handle_request()


if __name__ == '__main__':
    http_server()



"""
QEMU related utility functions.
"""
import re

from avocado.utils import process

QEMU_VERSION_RE = re.compile(r"QEMU (?:PC )?emulator version\s"
                             r"([0-9]+\.[0-9]+\.[0-9]+)"
                             r"(?:\s\((.*?)\))?")
DEVICE_CATEGORY_RE = re.compile(r"([A-Z]\S+) devices:")


def _get_info(bin_path, options, allow_output_check=None):
    """
    Execute a qemu command and return its stdout

    :param bin_path: Path to qemu binary
    :param options: Command line to run
    :param allow_output_check: Record the output from stdout/stderr
    :return: Command stdout
    """
    qemu_cmd = "%s %s" % (bin_path, options)
    return process.run(qemu_cmd, verbose=False, ignore_status=True,
                       allow_output_check=allow_output_check).stdout_text.strip()


def get_qemu_version(bin_path):
    """
    Return normalized qemu version and package version

    :param bin_path: Path to qemu binary
    :raise OSError: If unable to get that
    :return: A tuple of normalized version and package version
    """
    output = _get_info(bin_path, "-version")
    matches = QEMU_VERSION_RE.match(output)
    if matches is None:
        raise OSError('Unable to get the version of qemu')
    return matches.groups()


def get_machines_info(bin_path):
    """
    Return all machines information supported by qemu

    :param bin_path: Path to qemu binary
    :return: A dict of all machines
    """
    output = _get_info(bin_path, r"-machine help", allow_output_check="combined")
    machines = re.findall(r"^([a-z]\S+)\s+(.*)$", output, re.M)
    return dict(machines)


def get_supported_machines_list(bin_path):
    """
    Return all machines supported by qemu

    :param bin_path: Path to qemu binary
    :return: A list of all machines supported by qemu
    """
    return list(get_machines_info(bin_path).keys())


def get_devices_info(bin_path, category=None):
    """
    Return all devices information supported by qemu

    :param bin_path: Path to qemu binary
    :param category: device category (e.g. 'USB', 'Network', 'CPU')
    :return: A dict of all devices
    """
    qemu_devices = {}
    output = _get_info(bin_path, "-device help", allow_output_check="combined")
    require_machine = "No machine specified" in output
    # Some architectures (arm) require machine type to be always set, but this
    # function is not yet supported
    if require_machine:
        return qemu_devices

    for device_info in output.split("\n\n"):
        device_type = DEVICE_CATEGORY_RE.match(device_info)
        if device_type:
            device_type = device_type.group(1)
            devs_info = re.findall(r'^name "(\S+)"(.*)', device_info,
                                   re.M)
            qemu_devices[device_type] = {dev[0]: dev[1].replace(", ", "", 1)
                                         for dev in devs_info}
    if category:
        return qemu_devices.get(category, {})
    return {k: v for d in qemu_devices.values() for k, v in d.items()}


def get_supported_devices_list(bin_path, category=None):
    """
    Return all devices supported by qemu

    :param bin_path: Path to qemu binary
    :param category: device category (e.g. 'USB', 'Network', 'CPU')
    :return: A list of all devices supported by qemu
    """
    return list(get_devices_info(bin_path, category).keys())


def has_device_category(bin_path, category):
    """
    Check if device category is included in the qemu devices info

    :param bin_path: Path to qemu binary
    :param category: device category (e.g. 'USB', 'Network', 'CPU')
    :return: True if device category existed in qemu devices help info
    """
    out = _get_info(bin_path, "-device help", allow_output_check="combined")
    return category in DEVICE_CATEGORY_RE.findall(out)


def find_supported_devices(bin_path, pattern, category=None):
    """
    Use pattern to find all matching devices.

    :param bin_path: Path to qemu binary
    :param pattern: pattern to search the most suitable device
    :param category: device category (e.g. 'USB', 'Network', 'CPU')
    :return: device list
    """
    devices = []
    for device in get_supported_devices_list(bin_path, category):
        if re.match(pattern, device):
            devices.append(device)
    return devices


def get_maxcpus_hard_limit(bin_path, machine_type):
    """
    Return maximum limit CPUs supported by specified machine type

    :param bin_path: Path to qemu binary
    :param machine_type: One machine type supported by qemu
    :raise ValueError: If unable to get that
    :return: Maximum value of vCPU
    """
    invalid_maximum = 0x7fffffff
    output = _get_info(bin_path, r"-machine %s -smp maxcpus=%d"
                       % (machine_type, invalid_maximum),
                       allow_output_check="combined")
    searches = re.search(r"'%s.*' is (\d+)" % machine_type, output)
    if searches is None:
        raise ValueError("Could not get the maximum limit CPUs supported by "
                         "this machine '%s'" % machine_type)
    return int(searches.group(1))



from __future__ import division
import logging
import time
import glob
import os
import re
import socket
import traceback
import functools

from aexpect import remote
from aexpect.exceptions import ShellError
from aexpect.exceptions import ExpectError

from avocado.core import exceptions

import six
from six.moves import xrange

from virttest import utils_misc
from virttest import utils_net
from virttest import remote as remote_old
from virttest import ppm_utils
from virttest import data_dir
from virttest import error_context
from virttest import vt_console


class VMError(Exception):

    def __init__(self, *args):
        Exception.__init__(self, *args)


class VMUnexpectedExitError(VMError):

    def __init__(self, vm, exit_status=None):
        super(VMUnexpectedExitError, self).__init__(vm, exit_status)
        self.vm = vm
        self.exit_status = exit_status
        self.msg = None

    def __str__(self):
        if self.msg is None:
            self.msg = "vm %s is exited unexpectedly%s."
            if self.exit_status is None:
                self.msg %= (self.vm, "")
            else:
                self.msg %= (self.vm, ", exit status: %s" % self.exit_status)
        return self.msg


class VMCreateError(VMError):

    def __init__(self, cmd, status, output):
        VMError.__init__(self, cmd, status, output)
        self.cmd = cmd
        self.status = status
        self.output = output

    def __str__(self):
        return ("VM creation command failed:    %r    (status: %s,    "
                "output: %r)" % (self.cmd, self.status, self.output))


class VMStartError(VMError):

    def __init__(self, name, reason=None):
        VMError.__init__(self, name, reason)
        self.name = name
        self.reason = reason

    def __str__(self):
        msg = "VM '%s' failed to start" % self.name
        if self.reason is not None:
            msg += ": %s" % self.reason
        return msg


class VMConfigMissingError(VMError):

    def __init__(self, name, config):
        VMError.__init__(self, name, config)
        self.name = name
        self.config = config

    def __str__(self):
        return "Missing config '%s' for VM %s" % (self.config, self.name)


class VMHashMismatchError(VMError):

    def __init__(self, actual, expected):
        VMError.__init__(self, actual, expected)
        self.actual_hash = actual
        self.expected_hash = expected

    def __str__(self):
        return ("CD image hash (%s) differs from expected one (%s)" %
                (self.actual_hash, self.expected_hash))


class VMImageMissingError(VMError):

    def __init__(self, filename):
        VMError.__init__(self, filename)
        self.filename = filename

    def __str__(self):
        return "CD image file not found: %r" % self.filename


class VMImageCheckError(VMError):

    def __init__(self, filename):
        VMError.__init__(self, filename)
        self.filename = filename

    def __str__(self):
        return "Errors found on image: %r" % self.filename


class VMBadPATypeError(VMError):

    def __init__(self, pa_type):
        VMError.__init__(self, pa_type)
        self.pa_type = pa_type

    def __str__(self):
        return "Unsupported PCI assignable type: %r" % self.pa_type


class VMPAError(VMError):

    def __init__(self, pa_type):
        VMError.__init__(self, pa_type)
        self.pa_type = pa_type

    def __str__(self):
        return ("No PCI assignable devices could be assigned "
                "(pci_assignable=%r)" % self.pa_type)


class VMPostCreateError(VMError):

    def __init__(self, cmd, output):
        VMError.__init__(self, cmd, output)
        self.cmd = cmd
        self.output = output


class VMHugePageError(VMPostCreateError):

    def __str__(self):
        return ("Cannot allocate hugepage memory    (command: %r,    "
                "output: %r)" % (self.cmd, self.output))


class VMKVMInitError(VMPostCreateError):

    def __str__(self):
        return ("Cannot initialize KVM    (command: %r,    output: %r)" %
                (self.cmd, self.output))


class VMDeadError(VMError):

    def __init__(self, reason='', detail=''):
        VMError.__init__(self)
        self.reason = reason
        self.detail = detail

    def __str__(self):
        msg = "VM is dead"
        if self.reason:
            msg += "    reason: %s" % self.reason
        if self.detail:
            msg += "    detail: %r" % self.detail
        return (msg)


class VMDeadKernelCrashError(VMError):

    def __init__(self, kernel_crash):
        VMError.__init__(self, kernel_crash)
        logging.debug(kernel_crash)

    def __str__(self):
        return ("VM is dead due to a kernel crash, "
                "see debug/serial log for details")


class VMInvalidInstructionCode(VMError):

    def __init__(self, invalid_code):
        VMError.__init__(self, invalid_code)
        self.invalid_code = invalid_code

    def __str__(self):
        error = ""
        for invalid_code in self.invalid_code:
            error += "%s" % (invalid_code)
        return ("Invalid instruction was executed on VM:\n%s" % error)


class VMAddressError(VMError):
    pass


class VMInterfaceIndexError(VMError):
    pass


class VMPortNotRedirectedError(VMAddressError):

    def __init__(self, port, virtnet_nic=None):
        VMAddressError.__init__(self, port)
        self.port = port
        self.virtnet_nic = virtnet_nic

    def __str__(self):
        msg = "Don't know how to connect to guest port %s" % self.port
        if self.virtnet_nic is None:
            return msg
        else:
            nic = self.virtnet_nic
            msg += (" with networking type '%s', to destination '%s', for nic "
                    "'%s' with mac '%s' and ip '%s'." % (nic.nettype, nic.netdst,
                                                         nic.nic_name, nic.mac, nic.ip))
            return msg


class VMAddressVerificationError(VMAddressError):

    def __init__(self, mac, ip):
        VMAddressError.__init__(self, mac, ip)
        self.mac = mac
        self.ip = ip

    def __str__(self):
        return ("Could not verify DHCP lease: "
                "%s --> %s" % (self.mac, self.ip))


class VMMACAddressMissingError(VMAddressError):

    def __init__(self, nic_index):
        VMAddressError.__init__(self, nic_index)
        self.nic_index = nic_index

    def __str__(self):
        return "No MAC defined for NIC #%s" % self.nic_index


class VMIPAddressMissingError(VMAddressError):

    def __init__(self, mac, ip_version="ipv4"):
        VMAddressError.__init__(self, mac)
        self.mac = mac
        self.ip_version = ip_version

    def __str__(self):
        return "No %s DHCP lease for MAC %s" % (self.ip_version, self.mac)


class VMUnknownNetTypeError(VMError):

    def __init__(self, vmname, nicname, nettype):
        super(VMUnknownNetTypeError, self).__init__()
        self.vmname = vmname
        self.nicname = nicname
        self.nettype = nettype

    def __str__(self):
        return "Unknown nettype '%s' requested for NIC %s on VM %s" % (
            self.nettype, self.nicname, self.vmname)


class VMAddNetDevError(VMError):
    pass


class VMDelNetDevError(VMError):
    pass


class VMAddNicError(VMError):
    pass


class VMDelNicError(VMError):
    pass


class VMMigrateError(VMError):
    pass


class VMMigrateTimeoutError(VMMigrateError):
    pass


class VMMigrateCancelError(VMMigrateError):
    pass


class VMMigrateFailedError(VMMigrateError):
    pass


class VMMigrateProtoUnknownError(exceptions.TestSkipError):

    def __init__(self, protocol):
        self.protocol = protocol

    def __str__(self):
        return ("Virt Test doesn't know migration protocol '%s'. "
                "You would have to add it to the list of known protocols" %
                self.protocol)


class VMMigrateStateMismatchError(VMMigrateError):

    def __init__(self):
        VMMigrateError.__init__(self)

    def __str__(self):
        return ("Mismatch of VM state before and after migration")


class VMRebootError(VMError):
    pass


class VMStatusError(VMError):
    pass


class VMRemoveError(VMError):
    pass


class VMDeviceError(VMError):
    pass


class VMDeviceNotSupportedError(VMDeviceError):

    def __init__(self, name, device):
        VMDeviceError.__init__(self, name, device)
        self.name = name
        self.device = device

    def __str__(self):
        return ("Device '%s' is not supported for vm '%s' on this Host." %
                (self.device, self.name))


class VMDeviceCheckError(VMDeviceError):
    pass


class VMDeviceNotFoundError(VMDeviceError):
    pass


class VMDeviceStateError(VMDeviceError):
    pass


class VMPCIDeviceError(VMDeviceError):
    pass


class VMPCISlotInUseError(VMPCIDeviceError):

    def __init__(self, name, slot):
        VMPCIDeviceError.__init__(self, name, slot)
        self.name = name
        self.slot = slot

    def __str__(self):
        return ("PCI slot '0x%s' is already in use on vm '%s'. Please assign"
                " another slot in config file." % (self.slot, self.name))


class VMPCIOutOfRangeError(VMPCIDeviceError):

    def __init__(self, name, max_dev_num):
        VMPCIDeviceError.__init__(self, name, max_dev_num)
        self.name = name
        self.max_dev_num = max_dev_num

    def __str__(self):
        return ("Too many PCI devices added on vm '%s', max supported '%s'" %
                (self.name, str(self.max_dev_num)))


class VMUSBError(VMError):
    pass


class VMUSBControllerError(VMUSBError):
    pass


class VMUSBControllerMissingError(VMUSBControllerError):

    def __init__(self, name, controller_type):
        VMUSBControllerError.__init__(self, name, controller_type)
        self.name = name
        self.controller_type = controller_type

    def __str__(self):
        return ("Could not find '%s' USB Controller on vm '%s'. Please "
                "check config files." % (self.controller_type, self.name))


class VMUSBControllerPortFullError(VMUSBControllerError):

    def __init__(self, name, usb_dev_dict):
        VMUSBControllerError.__init__(self, name, usb_dev_dict)
        self.name = name
        self.usb_dev_dict = usb_dev_dict

    def __str__(self):
        output = ""
        try:
            for ctl, dev_list in six.iteritems(self.usb_dev_dict):
                output += "%s: %s\n" % (ctl, dev_list)
        except Exception:
            pass

        return ("No available USB port left on VM %s.\n"
                "USB devices map is: \n%s" % (self.name, output))


class VMUSBPortInUseError(VMUSBError):

    def __init__(self, vm_name, controller, port):
        VMUSBError.__init__(self, vm_name, controller, port)
        self.vm_name = vm_name
        self.controller = controller
        self.port = port

    def __str__(self):
        return ("USB port '%d' of controller '%s' is already in use on vm"
                " '%s'. Please assign another port in config file." %
                (self.port, self.controller, self.vm_name))


class VMScreenInactiveError(VMError):

    def __init__(self, vm, inactive_time):
        VMError.__init__(self)
        self.vm = vm
        self.inactive_time = inactive_time

    def __str__(self):
        msg = ("%s screen is inactive for %d s (%d min)" %
               (self.vm.name, self.inactive_time, self.inactive_time // 60))
        return msg


class VMLoginError(VMError):
    pass


class VMSMPTopologyInvalidError(VMError):
    pass


class CpuInfo(object):

    """
    A class for VM's cpu information.
    """

    def __init__(self, model=None, vendor=None, flags=None, family=None,
                 qemu_type=None, smp=0, maxcpus=0, cores=0, threads=0,
                 dies=0, sockets=0):
        """
        :param model: CPU Model of VM (use 'qemu -cpu ?' for list)
        :param vendor: CPU Vendor of VM
        :param flags: CPU Flags of VM
        :param family: CPU Family of VM
        :param qemu_type: cpu driver type of qemu
        :param smp: set the number of CPUs to 'n' [default=1]
        :param maxcpus: maximum number of total cpus, including
                        offline CPUs for hotplug, etc
        :param cores: number of CPU cores on one socket (for PC, it's on one die)
        :param threads: number of threads on one CPU core
        :param dies: number of CPU dies on one socket (for PC only)
        :param sockets: number of discrete sockets in the system
        """
        self.model = model
        self.vendor = vendor
        self.flags = flags
        self.family = family
        self.qemu_type = qemu_type
        self.smp = smp
        self.maxcpus = maxcpus
        self.cores = cores
        self.threads = threads
        self.dies = dies
        self.sockets = sockets


def session_handler(func):
    """
    decorator method to handle uri and session for libvirt
    """
    @functools.wraps(func)
    def manage_session(vm, *args, **kwargs):
        connect_uri = None
        uri = kwargs.get("connect_uri")
        libvirt = vm.params.get("vm_type") == 'libvirt'
        try:
            if uri and libvirt:
                connect_uri = vm.connect_uri
                vm.connect_uri = uri
                vm.session = vm.wait_for_serial_login()
            else:
                vm.session = vm.wait_for_login()
            return func(vm, *args, **kwargs)
        finally:
            if vm.session:
                vm.session.close()
            if connect_uri:
                vm.connect_uri = connect_uri
    return manage_session


class BaseVM(object):

    """
    Base class for all hypervisor specific VM subclasses.

    This class should not be used directly, that is, do not attempt to
    instantiate and use this class. Instead, one should implement a subclass
    that implements, at the very least, all methods defined right after the
    the comment blocks that are marked with:

    "Public API - *must* be reimplemented with virt specific code"

    and

    "Protected API - *must* be reimplemented with virt specific classes"

    The current proposal regarding methods naming convention is:

    - Public API methods: named in the usual way, consumed by tests
    - Protected API methods: name begins with a single underline, to be
      consumed only by BaseVM and subclasses
    - Private API methods: name begins with double underline, to be consumed
      only by the VM subclass itself (usually implements virt specific
      functionality)

    So called "protected" methods are intended to be used only by VM classes,
    and not be consumed by tests. Theses should respect a naming convention
    and always be preceded by a single underline.

    Currently most (if not all) methods are public and appears to be consumed
    by tests. It is a ongoing task to determine whether  methods should be
    "public" or "protected".
    """

    #
    # Assuming that all low-level hypervisor have at least migration via tcp
    # (true for xen & kvm). Also true for libvirt (using xen and kvm drivers)
    #
    MIGRATION_PROTOS = ['tcp', ]

    #
    # Timeout definition. This is being kept inside the base class so that
    # sub classes can change the default just for themselves
    #
    LOGIN_TIMEOUT = 10
    LOGIN_WAIT_TIMEOUT = 240
    COPY_FILES_TIMEOUT = 600
    MIGRATE_TIMEOUT = 3600
    REBOOT_TIMEOUT = 240

    def __init__(self, name, params):
        self.name = name
        self.params = params
        self.serial_console = None
        self.session = None
        # Create instance if not already set
        if not hasattr(self, 'instance'):
            self._generate_unique_id()
        # Don't overwrite existing state, update from params
        if hasattr(self, 'virtnet'):
            # Direct reference to self.virtnet makes pylint complain
            # note: virtnet.__init__() supports being called anytime
            getattr(self, 'virtnet').__init__(self.params,
                                              self.name,
                                              self.instance)
        else:  # Create new
            self.virtnet = utils_net.VirtNet(self.params,
                                             self.name,
                                             self.instance)
        self.ip_version = params.get("ip_version", "ipv4").lower()

        if not hasattr(self, 'cpuinfo'):
            self.cpuinfo = CpuInfo()
        if not hasattr(self, 'console_manager'):
            self.console_manager = vt_console.ConsoleManager()

    def _generate_unique_id(self):
        """
        Generate a unique identifier for this VM
        """
        while True:
            self.instance = (time.strftime("%Y%m%d-%H%M%S-") +
                             utils_misc.generate_random_string(8))
            if not glob.glob(os.path.join(data_dir.get_tmp_dir(),
                                          "*%s" % self.instance)):
                break

    def update_vm_id(self):
        """
        Update vm identifier, we need do that when force reboot vm, since vm
        virnet params may be changed.
        """
        self._generate_unique_id()

    @staticmethod
    def lookup_vm_class(vm_type, target):
        if vm_type == 'qemu':
            from virttest import qemu_vm
            return qemu_vm.VM
        if vm_type == 'libvirt':
            from virttest import libvirt_vm
            return libvirt_vm.VM
        if vm_type == 'v2v':
            if target == 'libvirt' or target is None:
                from virttest import libvirt_vm
                return libvirt_vm.VM
            if target == 'ovirt':
                from virttest import ovirt
                return ovirt.VMManager

    #
    # Public API - could be reimplemented with virt specific code
    #
    def needs_restart(self, name, params, basedir):
        """
        Verifies whether the current virt_install commandline matches the
        requested one, based on the test parameters.
        """
        if not self.is_alive():
            return True

        try:
            need_restart = (self.make_create_command() !=
                            self.make_create_command(name, params, basedir))
        except Exception:
            logging.error(traceback.format_exc())
            need_restart = True
        if need_restart:
            logging.debug(
                "VM params in env don't match requested, restarting.")
            return True
        else:
            # Command-line encoded state doesn't include all params
            # TODO: Check more than just networking
            other_virtnet = utils_net.VirtNet(params, name, self.instance)
            if self.virtnet != other_virtnet:
                logging.debug("VM params in env match, but network differs, "
                              "restarting")
                logging.debug("\t" + str(self.virtnet))
                logging.debug("\t!=")
                logging.debug("\t" + str(other_virtnet))
                return True
            else:
                logging.debug(
                    "VM params in env do match requested, continuing.")
                return False

    def verify_alive(self):
        """
        Make sure the VM is alive and that the main monitor is responsive.

        Can be subclassed to provide better information on why the VM is
        not alive (reason, detail)

        :raise VMDeadError: If the VM is dead
        :raise: Various monitor exceptions if the monitor is unresponsive
        """
        if self.is_dead():
            raise VMDeadError

    @session_handler
    def get_distro(self, connect_uri=None):
        """
        Get distribution name of the vm instance.
        """
        return utils_misc.get_distro(session=self.session)

    @session_handler
    def uptime(self, connect_uri=None):
        """
        Get uptime of the vm instance.

        :param connect_uri: Libvirt connect uri of vm
        :return: uptime of the vm on success, None on failure
        """
        return utils_misc.get_uptime(self.session)

    @session_handler
    def sosreport(self, path=None, connect_uri=None):
        """
        Get sosreport of the vm instance

        :param path: local host path where guest sosreport to be saved
        :param connect_uri: Connect uri for libvirt

        :return: host path where guest sosrepost saved, default to logdir
                 None if vm is not linux or sosreport fails.
        """
        log_path = None
        if not self.params["os_type"] == "linux":
            logging.warn("sosreport not applicable for %s",
                         self.params["os_type"])
            return None
        try:
            pkg = "sos"
            if "ubuntu" in self.get_distro().lower():
                pkg = "sosreport"
            guest_ip = self.get_address(session=self.session)
            guest_user = self.params["username"]
            guest_pwd = self.params["password"]
            log_path = utils_misc.get_sosreport(session=self.session,
                                                remote_ip=guest_ip,
                                                remote_pwd=guest_pwd,
                                                remote_user=guest_user,
                                                sosreport_name=self.name,
                                                sosreport_pkg=pkg)
        finally:
            return log_path

    def get_mac_address(self, nic_index=0):
        """
        Return the MAC address of a NIC.

        :param nic_index: Index of the NIC
        :return: MAC address of the NIC
        :raise VMMACAddressMissingError: If no MAC address is defined for the
                requested NIC
        """
        try:
            mac = self.virtnet[nic_index].mac
        except KeyError:
            raise VMMACAddressMissingError(nic_index)
        if mac is None:
            raise VMMACAddressMissingError(nic_index)
        return mac

    def get_address(self, index=0, ip_version="ipv4", session=None,
                    timeout=60.0):
        """
        Wrapper for self._get_address. if 'flexible_nic_index' is 'yes',
        will traverses from first nic to first element toward the end
        util get a reachable IP address;

        :param index: Name or index of the NIC whose address is requested.
        :param ip_version: IP version, value in 'ipv4' or 'ipv6,
        :param session: remote host session, if VM is migrated
        :param timeout: Timeout for retry verifying IP address and commands
        """
        nr_nics = len(self.virtnet.mac_list())
        nics_index = [index]
        flexible = self.params.get("flexible_nic_index") == "yes"
        if (flexible and nr_nics > 1):
            nics_index += [i for i in xrange(nr_nics) if i != index]

        for nic in nics_index:
            try:
                return self._get_address(nic, ip_version, session=session,
                                         timeout=timeout)
            except (VMMACAddressMissingError, VMIPAddressMissingError,
                    VMAddressVerificationError):
                if nic == nics_index[-1]:
                    raise

    def _get_address(self, index=0, ip_version="ipv4", session=None,
                     timeout=60.0):
        """
        Return the IP address of a NIC or guest (in host space).

        :param index: Name or index of the NIC whose address is requested.
        :param ip_version: IP version, value in 'ipv4' or 'ipv6,
                           default value is 'ipv4'
        :param session: ShellSession object of remote host
        :param timeout: Timeout for retry verifying IP address and commands
        :return: 'localhost': Port redirection is in use
        :return: IP address of NIC if valid in arp cache.
        :raise VMMACAddressMissingError: If no MAC address is defined for the
                requested NIC
        :raise VMIPAddressMissingError: If no IP address is found for the the
                NIC's MAC address
        :raise VMAddressVerificationError: If the MAC-IP address mapping cannot
                be verified
        """
        nic = self.virtnet[index]

        # TODO: Determine port redirection in use w/o checking nettype
        if nic.nettype not in ['bridge', 'macvtap']:
            hostname = socket.gethostname()
            if session:
                hostname = session.cmd_output("hostname -f", timeout=timeout)
            return socket.gethostbyname(hostname)

        mac = self.get_mac_address(index).lower()
        if (self.params.get('using_linklocal') == "yes" and
                ip_version == "ipv6"):
            return utils_net.ipv6_from_mac_addr(mac)

        # TODO: Redesign address_cache and declare it in this class
        mac_pattern = "%s_6" if ip_version == "ipv6" else "%s"
        ip_addr = self.address_cache.get(mac_pattern % mac)
        if not ip_addr:
            raise VMIPAddressMissingError(mac, ip_version)

        devs = set([nic.netdst]) if 'netdst' in nic else set()
        if not utils_net.verify_ip_address_ownership(ip_addr, [mac],
                                                     devs=devs,
                                                     session=session,
                                                     timeout=timeout):
            nic_params = self.params.object_params(nic.nic_name)
            pci_assignable = nic_params.get("pci_assignable") != "no"

            if not (pci_assignable or nic.nettype == "macvtap"):
                raise VMAddressVerificationError(mac, ip_addr)

            self.address_cache.drop(mac_pattern % mac)

            # SR-IOV/Macvtap cards may not be in same subnet with the cards
            # used by host by default, so arp checks won't work. Therefore,
            # do not raise VMAddressVerificationError when SR-IOV is used.
            nic_backend = pci_assignable and "SR-IOV" or "macvtap"
            msg = "Could not verify DHCP lease: %s-> %s." % (mac, ip_addr)
            msg += " Maybe %s is not in the same subnet " % ip_addr
            msg += "as the host (%s in use)" % nic_backend
            logging.error(msg)

        return ip_addr

    def fill_addrs(self, addrs):
        """
        Fill VM's nic address to the virtnet structure based on VM's address
        structure addrs.

        :param addrs: Dict of interfaces and address

        ::

            {"if_name":{"mac":['addrs',],
                        "ipv4":['addrs',],
                        "ipv6":['addrs',]},
              ...}
        """
        for virtnet in self.virtnet:
            for iface_name, iface in six.iteritems(addrs):
                if virtnet.mac in iface["mac"]:
                    virtnet.ip = {"ipv4": iface["ipv4"],
                                  "ipv6": iface["ipv6"]}
                    virtnet.g_nic_name = iface_name

    def get_port(self, port, nic_index=0):
        """
        Return the port in host space corresponding to port in guest space.

        :param port: Port number in host space.
        :param nic_index: Index of the NIC.
        :return: If port redirection is used, return the host port redirected
                to guest port port. Otherwise return port.
        :raise VMPortNotRedirectedError: If an unredirected port is requested
                in user mode
        """
        nic_nettype = self.virtnet[nic_index].nettype
        if nic_nettype in ["bridge", "macvtap"]:
            return port
        else:
            try:
                return self.redirs[port]
            except KeyError:
                raise VMPortNotRedirectedError(port, self.virtnet[nic_index])

    def free_mac_address(self, nic_index_or_name=0):
        """
        Free a NIC's MAC address.

        :param nic_index: Index of the NIC
        """
        self.virtnet.free_mac_address(nic_index_or_name)

    @error_context.context_aware
    def wait_for_get_address(self, nic_index, timeout=30,
                             interval=0.5, ip_version='ipv4'):
        """
        Wait for a nic to acquire an IP address, then return it.
        """
        # Don't let VMIPAddressMissingError/VMAddressVerificationError through
        def _get_address():
            try:
                return self.get_address(nic_index, ip_version)
            except (VMIPAddressMissingError, VMAddressVerificationError) as e:
                return False

        ipaddr = utils_misc.wait_for(_get_address, timeout, step=interval)
        if not ipaddr:
            # Read guest address via serial console and update VM address
            # cache to avoid get out-dated address.
            utils_net.update_mac_ip_address(self, timeout)
            ipaddr = self.get_address(nic_index, ip_version)
        msg = 'Found/Verified IP %s for VM %s NIC %s' % (ipaddr,
                                                         self.name,
                                                         nic_index)
        logging.debug(msg)
        return ipaddr

    # Adding/setup networking devices methods split between 'add_*' for
    # setting up virtnet, and 'activate_' for performing actions based
    # on settings.
    def add_nic(self, **params):
        """
        Add new or setup existing NIC with optional model type and mac address

        :param params: Dict with additional NIC parameters to set.
        :return: Dict with new NIC's info.
        """
        if 'nic_name' not in params:
            params['nic_name'] = utils_misc.generate_random_id()
        nic_name = params['nic_name']
        if nic_name in self.virtnet.nic_name_list():
            self.virtnet[nic_name].update(**params)
        else:
            self.virtnet.append(params)
        nic = self.virtnet[nic_name]
        if 'mac' not in nic:  # generate random mac
            logging.debug("Generating random mac address for nic")
            self.virtnet.generate_mac_address(nic_name)
        # mac of '' or invaid format results in not setting a mac
        if 'ip' in nic and 'mac' in nic:
            self.address_cache[nic.mac] = nic.ip
        return nic

    def del_nic(self, nic_index_or_name):
        """
        Remove the nic specified by name, or index number
        """
        nic = self.virtnet[nic_index_or_name]
        nic_mac = nic.mac.lower()
        self.free_mac_address(nic_index_or_name)
        try:
            del self.virtnet[nic_index_or_name]
            self.address_cache.drop(nic_mac)
        except IndexError:
            pass  # continue to not exist
        except KeyError:
            pass  # continue to not exist

    def get_nic_index_by_mac(self, mac):
        """
        Get the nic index specified by MAC address

        :param mac: MAC address
        :return: nic index number
        """
        for index, nic in enumerate(self.virtnet):
            if 'mac' not in nic:
                continue
            elif nic.mac == mac:
                return index
        logging.warn("Not find nic by '%s'", mac)
        return -1

    def verify_kernel_crash(self):
        """
        Find kernel crash message on the VM serial console.

        :raise: VMDeadKernelCrashError, in case a kernel crash message was
                found.
        """
        panic_re = [r"BUG:.*---\[ end trace .* \]---"]
        panic_re.append(r"----------\[ cut here.* BUG .*\[ end trace .* \]---")
        panic_re.append(r"general protection fault:.* RSP.*>")
        panic_re = "|".join(panic_re)
        if self.serial_console:
            data = self.serial_console.get_output()
            if data is None:
                logging.warn("Unable to read serial console")
                return
            match = re.search(panic_re, data, re.DOTALL | re.MULTILINE | re.I)
            if match:
                raise VMDeadKernelCrashError(match.group(0))

    @session_handler
    def verify_dmesg(self, dmesg_log_file=None, connect_uri=None):
        """
        Verify guest dmesg

        :param dmesg_log_file: The file used to save guest dmesg. If None, will save
                               guest dmesg to logging.debug.
        :param connect_uri: Libvirt connect uri of vm
        """
        level = self.params.get("guest_dmesg_level", 3)
        ignore_result = self.params.get("guest_dmesg_ignore", "no") == "yes"
        serial_login = self.params.get("serial_login", "no") == "yes"
        if serial_login:
            self.session = self.wait_for_serial_login()
        elif(len(self.virtnet) > 0 and self.virtnet[0].nettype != "macvtap" and
             not connect_uri):
            self.session = self.wait_for_login()
        return utils_misc.verify_dmesg(dmesg_log_file=dmesg_log_file,
                                       ignore_result=ignore_result,
                                       level_check=level,
                                       session=self.session)

    def verify_bsod(self, scrdump_file):
        # For windows guest
        if (os.path.exists(scrdump_file) and
                self.params.get("check_guest_bsod", "no") == 'yes' and
                ppm_utils.Image is not None):
            ref_img_path = self.params.get("bsod_reference_img", "")
            bsod_base_dir = os.path.join(data_dir.get_root_dir(),
                                         "shared", "deps",
                                         "bsod_img")
            ref_img = utils_misc.get_path(bsod_base_dir, ref_img_path)
            if ppm_utils.have_similar_img(scrdump_file, ref_img):
                err_msg = "Windows Guest appears to have suffered a BSOD,"
                err_msg += " please check %s against %s." % (
                    scrdump_file, ref_img)
                raise VMDeadKernelCrashError(err_msg)

    def verify_illegal_instruction(self):
        """
        Find illegal instruction code on VM serial console output.

        :raise: VMInvalidInstructionCode, in case a wrong instruction code.
        """
        if self.serial_console is not None:
            data = self.serial_console.get_output()
            if data is None:
                logging.warn("Unable to read serial console")
                return
            match = re.findall(r".*trap invalid opcode.*\n", data,
                               re.MULTILINE)

            if match:
                raise VMInvalidInstructionCode(match)

    def get_params(self):
        """
        Return the VM's params dict. Most modified params take effect only
        upon VM.create().
        """
        return self.params

    def get_testlog_filename(self):
        """
        Return the testlog filename.
        """
        return os.path.join(data_dir.get_tmp_dir(),
                            "testlog-%s" % self.instance)

    @error_context.context_aware
    def login(self, nic_index=0, timeout=LOGIN_TIMEOUT,
              username=None, password=None):
        """
        Log into the guest via SSH/Telnet/Netcat.
        If timeout expires while waiting for output from the guest (e.g. a
        password prompt or a shell prompt) -- fail.

        :param nic_index: The index of the NIC to connect to.
        :param timeout: Time (seconds) before giving up logging into the
                guest.
        :return: A ShellSession object.
        """
        error_context.context("logging into '%s'" % self.name)
        if not username:
            username = self.params.get("username", "")
        if not password:
            password = self.params.get("password", "")
        prompt = self.params.get("shell_prompt", r"[\#\$]\s*$")
        linesep = eval("'%s'" % self.params.get("shell_linesep", r"\n"))
        client = self.params.get("shell_client")
        try:
            address = self.get_address(nic_index, self.ip_version)
        except (VMIPAddressMissingError, VMAddressVerificationError) as e:
            utils_net.update_mac_ip_address(self, timeout)
            address = self.get_address(nic_index, self.ip_version)
        neigh_attach_if = ""
        if self.ip_version == "ipv6" and address.lower().startswith("fe80"):
            neigh_attach_if = utils_net.get_neigh_attch_interface(address)
        port = self.get_port(int(self.params.get("shell_port")))
        log_filename = ("session-%s-%s-%s.log" %
                        (self.name, time.strftime("%m-%d-%H-%M-%S"),
                         utils_misc.generate_random_string(4)))
        session = remote.remote_login(client, address, port, username,
                                      password, prompt, linesep,
                                      log_filename, timeout,
                                      interface=neigh_attach_if)
        session.set_status_test_command(self.params.get("status_test_command",
                                                        ""))
        self.remote_sessions.append(session)
        return session

    @error_context.context_aware
    def commander(self, nic_index=0, timeout=LOGIN_TIMEOUT,
                  username=None, password=None, commander_path=None):
        """
        Log into the guest via SSH/Telnet/Netcat.
        If timeout expires while waiting for output from the guest (e.g. a
        password prompt or a shell prompt) -- fail.

        :param nic_index: The index of the NIC to connect to.
        :param timeout: Time (seconds) before giving up logging into the
                guest.
        :param commander_path: Path where will be commander placed.
        :return: A ShellSession object.
        """
        if commander_path is None:
            commander_path = data_dir.get_tmp_dir()
        error_context.context("logging into '%s'" % self.name)
        if not username:
            username = self.params.get("username", "")
        if not password:
            password = self.params.get("password", "")
        prompt = "^\s*#"
        linesep = eval("'%s'" % self.params.get("shell_linesep", r"\n"))
        client = self.params.get("shell_client")
        address = self.get_address(nic_index)
        port = self.get_port(int(self.params.get("shell_port")))
        log_filename = None

        from virttest import remote_commander as rc
        path = os.path.dirname(rc.__file__)
        f_path = [os.path.join(path, _) for _ in
                  ("remote_runner.py", "remote_interface.py", "messenger.py")]
        self.copy_files_to(f_path, commander_path)

        # start remote commander
        cmd = remote_old.remote_commander(client, address, port, username,
                                          password, prompt, linesep, log_filename,
                                          timeout, commander_path)
        self.remote_sessions.append(cmd)
        return cmd

    def wait_for_login(self, nic_index=0, timeout=LOGIN_WAIT_TIMEOUT,
                       internal_timeout=LOGIN_TIMEOUT,
                       serial=False, restart_network=False,
                       username=None, password=None, status_check=True):
        """
        Make multiple attempts to log into the guest via SSH/Telnet/Netcat.

        :param nic_index: The index of the NIC to connect to.
        :param timeout: Time (seconds) to keep trying to log in.
        :param internal_timeout: Timeout to pass to login().
        :param serial: Whether to use a serial connection when remote login
                (ssh, rss) failed.
        :param restart_network: Whether to try to restart guest's network
                when remote login (ssh, rss) failed.
        :param status_check: Whether to call verify_alive to detect bad
            VM state early. Disable this when VM status might be unreliable,
            eg. during reboot or pause)
        :return: A ShellSession object.
        """
        def print_guest_network_info():
            """
            Print guest network information into debug log file
            """
            session = None
            try:
                session = self.serial_login(internal_timeout, username,
                                            password)
                out = session.cmd_output("ipconfig || ifconfig", timeout=60)
                txt = ["Guest network status:\n %s" % out]
                out = session.cmd_output("ip route || route print", timeout=60)
                txt += ["Guest route table:\n %s" % out]
                logging.error("\n".join(txt))
            except Exception as e:
                logging.error("Can't get guest network status "
                              "information, reason: %s", e)
            finally:
                if session:
                    session.close()

        error = None
        logging.debug("Attempting to log into '%s' (timeout %ds)",
                      self.name, timeout)
        start_time = time.time()
        try:
            self.wait_for_get_address(nic_index,
                                      timeout=timeout,
                                      ip_version=self.ip_version)
        except Exception as err:
            error = err
            if status_check:
                self.verify_alive()
            print_guest_network_info()
            if not (serial or restart_network):
                raise
            session = self.wait_for_serial_login(timeout, internal_timeout,
                                                 restart_network, username,
                                                 password, False)
            if serial:
                return session
            session.close()

        # try to login if VM bootup really, at least once
        not_tried = True
        end_time = start_time + timeout
        while time.time() < end_time or not_tried:
            try:
                return self.login(nic_index, internal_timeout,
                                  username, password)
            except (remote.LoginAuthenticationError,
                    remote.LoginBadClientError):
                if serial:
                    break
                raise
            except Exception as err:
                time.sleep(0.5)
                error = err
            not_tried = False

        print_guest_network_info()
        if serial:
            return self.wait_for_serial_login(timeout, internal_timeout,
                                              False, username, password)

        raise remote.LoginTimeoutError("exceeded %s s timeout, last "
                                       "failure: %s" % (timeout, error))

    @error_context.context_aware
    def copy_files_to(self, host_path, guest_path, nic_index=0, limit="",
                      verbose=False, timeout=COPY_FILES_TIMEOUT,
                      username=None, password=None, filesize=None):
        """
        Transfer files to the remote host(guest).

        :param host_path: Host path
        :param guest_path: Guest path
        :param nic_index: The index of the NIC to connect to.
        :param limit: Speed limit of file transfer.
        :param verbose: If True, log some stats using logging.debug (RSS only)
        :param timeout: Time (seconds) before giving up on doing the remote
                copy.
        :param filesize: size of file will be transferred
        """
        error_context.context("sending file(s) to '%s'" % self.name)
        if not username:
            username = self.params.get("username", "")
        if not password:
            password = self.params.get("password", "")
        client = self.params.get("file_transfer_client")
        address = self.get_address(nic_index)
        neigh_attach_if = ""
        if self.ip_version == "ipv6" and address.lower().startswith("fe80"):
            neigh_attach_if = utils_net.get_neigh_attch_interface(address)
        port = self.get_port(int(self.params.get("file_transfer_port")))
        log_filename = ("transfer-%s-to-%s-%s.log" %
                        (self.name, address,
                         utils_misc.generate_random_string(4)))
        remote.copy_files_to(address, client, username, password, port,
                             host_path, guest_path, limit, log_filename,
                             verbose, timeout, interface=neigh_attach_if,
                             filesize=filesize)
        utils_misc.close_log_file(log_filename)

    @error_context.context_aware
    def copy_files_from(self, guest_path, host_path, nic_index=0, limit="",
                        verbose=False, timeout=COPY_FILES_TIMEOUT,
                        username=None, password=None, filesize=None):
        """
        Transfer files from the guest.

        :param host_path: Guest path
        :param guest_path: Host path
        :param nic_index: The index of the NIC to connect to.
        :param limit: Speed limit of file transfer.
        :param verbose: If True, log some stats using logging.debug (RSS only)
        :param timeout: Time (seconds) before giving up on doing the remote
                copy.
        :param filesize: size of file will be transferred
        """
        error_context.context("receiving file(s) from '%s'" % self.name)
        if not username:
            username = self.params.get("username", "")
        if not password:
            password = self.params.get("password", "")
        client = self.params.get("file_transfer_client")
        address = self.get_address(nic_index)
        neigh_attach_if = ""
        if self.ip_version == "ipv6" and address.lower().startswith("fe80"):
            neigh_attach_if = utils_net.get_neigh_attch_interface(address)
        port = self.get_port(int(self.params.get("file_transfer_port")))
        log_filename = ("transfer-%s-from-%s-%s.log" %
                        (self.name, address,
                         utils_misc.generate_random_string(4)))
        remote.copy_files_from(address, client, username, password, port,
                               guest_path, host_path, limit, log_filename,
                               verbose, timeout, interface=neigh_attach_if,
                               filesize=filesize)
        utils_misc.close_log_file(log_filename)

    def _create_serial_console(self):
        """
        Establish a session with the serial console.

        Let's consider the first serial port as serial console.
        Note: requires a version of netcat that supports -U
        """
        raise NotImplementedError

    def create_serial_console(self):
        """A Wrapper of _create_serial_console."""
        self._create_serial_console()
        self.console_manager.set_console(self.serial_console)

    def cleanup_serial_console(self):
        """
        Close serial console and associated log file
        """
        raise NotImplementedError

    def create_virtio_console(self):
        """
        Establish a session with the virtio console.
        """
        raise NotImplementedError

    @error_context.context_aware
    def serial_login(self, timeout=LOGIN_TIMEOUT,
                     username=None, password=None, virtio=False):
        """
        Log into the guest via the serial console.
        If timeout expires while waiting for output from the guest (e.g. a
        password prompt or a shell prompt) -- fail.

        :param timeout: Time (seconds) before giving up logging into the guest.
        :param virtio: is a console virtio console (deprecated).
        :return: ConsoleSession instance.
        """
        error_context.context("Logging into '%s' via serial console." %
                              self.name)
        if not username:
            username = self.params.get("username", "")
        if not password:
            password = self.params.get("password", "")

        prompt = self.params.get("shell_prompt", r"[\#\$]\s*$")
        linesep = eval("'%s'" % self.params.get("shell_linesep", r"\n"))
        status_test_command = self.params.get("status_test_command", "")

        return self.console_manager.create_session(linesep,
                                                   status_test_command,
                                                   prompt,
                                                   username,
                                                   password,
                                                   timeout)

    def wait_for_serial_login(self, timeout=LOGIN_WAIT_TIMEOUT,
                              internal_timeout=LOGIN_TIMEOUT,
                              restart_network=False,
                              username=None, password=None, virtio=False,
                              status_check=True):
        """
        Make multiple attempts to log into the guest via serial console.

        :param timeout: Time (seconds) to keep trying to log in.
        :param internal_timeout: Timeout to pass to serial_login().
        :param restart_network: Whether try to restart guest's network.
        :param virtio: is a console virtio console (deprecated).
        :param status_check: Whether to call verify_alive to detect bad
            VM state early. Disable this when VM status might be unreliable,
            eg. during reboot or pause)
        :return: ConsoleSession instance.
        """
        logging.debug("Attempting to log into '%s' via serial console "
                      "(timeout %ds)", self.name, timeout)
        end_time = time.time() + timeout
        while time.time() < end_time:
            try:
                session = self.serial_login(internal_timeout,
                                            username,
                                            password,
                                            virtio=virtio)
                break
            except remote.LoginError:
                if status_check:
                    self.verify_alive()
                time.sleep(0.5)
                continue
        else:
            raise remote.LoginTimeoutError('exceeded %s s timeout' %
                                           timeout)
        if restart_network:
            try:
                logging.debug("Attempting to restart guest network")
                os_type = self.params.get('os_type')
                utils_net.restart_guest_network(session, os_type=os_type)
            except (ShellError, ExpectError):
                session.close()
                raise
        return session

    def get_uuid(self):
        """
        Catch UUID of the VM.

        :return: None,if not specified in config file
        """
        if self.params.get("uuid") == "random":
            return self.uuid
        else:
            return self.params.get("uuid", None)

    def send_string(self, sr):
        """
        Send a string to the VM.

        :param sr: String, that must consist of alphanumeric characters only.
                Capital letters are allowed.
        """
        for char in sr:
            if char.isupper():
                self.send_key("shift-%s" % char.lower())
            else:
                self.send_key(char)

    @session_handler
    def get_cpu_count(self, check_cmd='cpu_chk_cmd', connect_uri=None):
        """
        Get the cpu count of the VM.
        """
        cmd = self.params.get(check_cmd)
        out = self.session.cmd_output_safe(cmd)
        return int(re.search("\d+", out, re.M).group())

    def get_memory_size(self, cmd=None, timeout=60):
        """
        Get bootup memory size of the VM.

        :param cmd: Command used to check memory. If not provided,
                    self.params.get("mem_chk_cmd") will be used.
        :param timeout: timeout for cmd
        """
        if not cmd:
            cmd = self.params["mem_chk_cmd"]
        session = self.wait_for_login()
        try:
            output = session.cmd_output(cmd, timeout=timeout).replace(',', '')
            mem_size = re.findall(r"\d+\s*[BbKkMmGgTt]?", output, re.M)
            num_mem_size = map(utils_misc.normalize_data_size, mem_size)
            return int(sum(map(float, num_mem_size)))
        finally:
            session.close()

    def get_current_memory_size(self):
        """
        Get current memory size of the VM, rather than bootup memory.
        """
        cmd = self.params.get("mem_chk_cur_cmd")
        return self.get_memory_size(cmd)

    def get_totalmem_sys(self, online='yes', node=''):
        """
        To get the total guest memory(ram) as detected by system
        MemTotal in /proc/meminfo would display
        total usable memory(i.e. physical ram minus
        a few reserved bits and the kernel binary code)
        :param online: if 'yes', count the total online memory size
        :param node: if given will count mem on that numa node
        :return: system memory in Kb as float
        """
        session = self.wait_for_login()
        try:
            if node != '':
                cmd = "count=0;[ -d /sys/devices/system/node/node%s/ ] && " % node
                cmd += "cd /sys/devices/system/node/node%s/;" % node
                cmd += "for i in `ls -d memory*`;"
            else:
                cmd = "count=0;cd /sys/devices/system/memory/;for i in `ls -d memory*`;"
            if online == 'yes':
                cmd += "do [ -f $i/online ] && a=$(<$i/online) && "
            else:
                cmd += "do [ -f $i/online ] && a=1 && "
            cmd += "count=$(( $count + $a ));a=0;done;echo $count"
            output = session.cmd_status_output(cmd, timeout=360)
            # Handle memory less numa nodes
            if "ls: cannot access 'memory*':" in output[1]:
                no_memblocks = 0
            else:
                no_memblocks = int(output[1])
            cmd = "cat /sys/devices/system/memory/block_size_bytes"
            block_size = int(session.cmd_output(cmd), 16)
            return (no_memblocks * block_size)/1024.0
        finally:
            session.close()

    #
    # Public API - *must* be reimplemented with virt specific code
    #
    def is_alive(self):
        """
        Return True if the VM is alive and the management interface is responsive.
        """
        raise NotImplementedError

    def is_dead(self):
        """
        Return True if the VM is dead.
        """
        raise NotImplementedError

    def is_paused(self):
        """
        Return True if the VM is paused
        """
        raise NotImplementedError

    def activate_nic(self, nic_index_or_name):
        """
        Activate an inactive network device

        :param nic_index_or_name: name or index number for existing NIC
        """
        raise NotImplementedError

    def deactivate_nic(self, nic_index_or_name):
        """
        Deactivate an active network device

        :param nic_index_or_name: name or index number for existing NIC
        """
        raise NotImplementedError

    def verify_userspace_crash(self):
        """
        Verify if the userspace component of the virtualization backend crashed.
        """
        pass

    def clone(self, name, **params):
        """
        Return a clone of the VM object with optionally modified parameters.

        This method should be implemented by
        """
        raise NotImplementedError

    def destroy(self, gracefully=True, free_mac_addresses=True):
        """
        Destroy the VM.

        If gracefully is True, first attempt to shutdown the VM with a shell
        command.  Then, attempt to destroy the VM via the monitor with a 'quit'
        command.  If that fails, send SIGKILL to the qemu process.

        :param gracefully: If True, an attempt will be made to end the VM
                using a shell command before trying to end the qemu process
                with a 'quit' or a kill signal.
        :param free_mac_addresses: If True, the MAC addresses used by the VM
                will be freed.
        """
        raise NotImplementedError

    def migrate(self, timeout=MIGRATE_TIMEOUT, protocol="tcp",
                cancel_delay=None, offline=False, stable_check=False,
                clean=True, save_path=data_dir.get_tmp_dir(),
                dest_host="localhost",
                remote_port=None):
        """
        Migrate the VM.

        If the migration is local, the VM object's state is switched with that
        of the destination VM.  Otherwise, the state is switched with that of
        a dead VM (returned by self.clone()).

        :param timeout: Time to wait for migration to complete.
        :param protocol: Migration protocol ('tcp', 'unix' or 'exec').
        :param cancel_delay: If provided, specifies a time duration after which
                migration will be canceled.  Used for testing migrate_cancel.
        :param offline: If True, pause the source VM before migration.
        :param stable_check: If True, compare the VM's state after migration to
                its state before migration and raise an exception if they
                differ.
        :param clean: If True, delete the saved state files (relevant only if
                stable_check is also True).
        :param save_path: The path for state files.
        :param dest_host: Destination host (defaults to 'localhost').
        :param remote_port: Port to use for remote migration.
        :param mig_inner_funcs: Functions to be executed just after the migration
                is started.
        """
        raise NotImplementedError

    def reboot(self, session=None, method="shell", nic_index=0,
               timeout=REBOOT_TIMEOUT, serial=False):
        """
        Reboot the VM and wait for it to come back up by trying to log in until
        timeout expires.

        :param session: A shell session object or None.
        :param method: Reboot method.  Can be "shell" (send a shell reboot
                command) or "system_reset" (send a system_reset monitor command).
        :param nic_index: Index of NIC to access in the VM, when logging in
                after rebooting.
        :param timeout: Time to wait for login to succeed (after rebooting).
        :param serial: Serial port login or not (default is False).
        :return: A new shell session object.
        """
        raise NotImplementedError

    # should this really be expected from VMs of all hypervisor types?
    def send_key(self, keystr):
        """
        Send a key event to the VM.

        :param keystr: A key event string (e.g. "ctrl-alt-delete")
        """
        raise NotImplementedError

    def save_to_file(self, path):
        """
        State of paused VM recorded to path and VM shutdown on success

        Throws a VMStatusError if before/after state is incorrect.

        :param path: file where VM state recorded

        """
        raise NotImplementedError

    def restore_from_file(self, path):
        """
        A shutdown or paused VM is resumed from path, & possibly set running

        Throws a VMStatusError if before/after restore state is incorrect

        :param path: path to file vm state was saved to
        """
        raise NotImplementedError

    def savevm(self, tag_name):
        """
        Save the virtual machine as the tag 'tag_name'

        :param tag_name: tag of the virtual machine that saved

        """
        raise NotImplementedError

    def loadvm(self, tag_name):
        """
        Load the virtual machine tagged 'tag_name'.

        :param tag_name: tag of the virtual machine that saved
        """
        raise NotImplementedError

    def pause(self):
        """
        Stop the VM operation.
        """
        raise NotImplementedError

    def resume(self):
        """
        Resume the VM operation in case it's stopped.
        """
        raise NotImplementedError




"""
Virsh net* command related utility functions
"""

from avocado.core import exceptions

from virttest import virsh
from virttest import remote
from virttest.utils_test import libvirt


def create_or_del_network(net_dict, is_del=False, remote_args=None):
    """
    Create or delete network on local or remote

    :param net_dict: Dictionary with the network parameters
    :param is_del: Whether the networks should be deleted
    :param remote_args: The parameters for remote
    """

    remote_virsh_session = None
    if remote_args:
        remote_virsh_session = virsh.VirshPersistent(**remote_args)

    if not is_del:
        net_dev = libvirt.create_net_xml(net_dict.get("net_name"), net_dict)

        if not remote_virsh_session:
            if net_dev.get_active():
                net_dev.undefine()
            net_dev.define()
            net_dev.start()
        else:
            remote_ip = remote_args.get("remote_ip")
            remote_user = remote_args.get("remote_user")
            remote_pwd = remote_args.get("remote_pwd")
            if not all([remote_ip, remote_user, remote_pwd]):
                raise exceptions.TestError("remote_[ip|user|pwd] are necessary!")
            remote.scp_to_remote(remote_ip, '22', remote_user, remote_pwd,
                                 net_dev.xml, net_dev.xml, limit="",
                                 log_filename=None, timeout=600,
                                 interface=None)
            remote_virsh_session.net_define(net_dev.xml, debug=True)
            remote_virsh_session.net_start(net_dict.get("net_name"),
                                           debug=True)
            remote.run_remote_cmd("rm -rf %s" % net_dev.xml, remote_args)
    else:
        virsh_session = virsh
        if remote_virsh_session:
            virsh_session = remote_virsh_session
        if net_dict.get("net_name") in virsh_session.net_state_dict():
            virsh_session.net_destroy(net_dict.get("net_name"),
                                      debug=True, ignore_status=True)
            virsh_session.net_undefine(net_dict.get("net_name"),
                                       debug=True, ignore_status=True)
    if remote_virsh_session:
        remote_virsh_session.close_session()



"""
Module to hide underlying device xml handler class implementation
"""

import os

from virttest.libvirt_xml import base


# Avoid accidental names like __init__, librarian, and/or other support modules
DEVICE_TYPES = ['disk', 'filesystem', 'controller', 'lease',
                'hostdev', 'redirdev', 'smartcard', 'interface', 'input',
                'hub', 'graphics', 'video', 'parallel', 'serial', 'console',
                'channel', 'sound', 'watchdog', 'memballoon', 'rng', 'vsock',
                'seclabel', 'address', 'emulator', 'panic', 'memory', 'filterref',
                'iommu', 'tpm']


def get(name):
    """
    Returns named device xml element's handler class

    :param name: the device name
    :return: the named device xml element's handler class
    """
    mod_path = os.path.abspath(os.path.dirname(__file__))
    handler_cl = base.load_xml_module(mod_path, name, DEVICE_TYPES)
    return handler_cl



import os
import re
import logging
from tempfile import mktemp

from avocado.core import exceptions
from avocado.utils import process

from virttest import virsh
from virttest import utils_misc
from virttest.utils_test import libvirt
from virttest.libvirt_xml.nodedev_xml import NodedevXML
from virttest.libvirt_xml.devices import hostdev


_FC_HOST_PATH = "/sys/class/fc_host"
_TIMEOUT = 5


def check_nodedev(dev_name, dev_parent=None):
    """
    Check node device relevant values

    :params dev_name: name of the device
    :params dev_parent: parent name of the device, None is default
    :return: True if nodedev is normal.
    """
    host = dev_name.split("_")[1]
    fc_host_path = os.path.join(_FC_HOST_PATH, host)

    # Check if the /sys/class/fc_host/host$NUM exists
    if not os.access(fc_host_path, os.R_OK):
        logging.error("Can't access %s", fc_host_path)
        return False

    dev_xml = NodedevXML.new_from_dumpxml(dev_name)
    if not dev_xml:
        logging.error("Can't dumpxml %s XML", dev_name)
        return False

    # Check device parent name
    if dev_parent != dev_xml.parent:
        logging.error("The parent name is different: %s is not %s",
                      dev_parent, dev_xml.parent)
        return False

    wwnn_from_xml = dev_xml.wwnn
    wwpn_from_xml = dev_xml.wwpn
    fabric_wwn_from_xml = dev_xml.fabric_wwn

    fc_dict = {}
    name_list = ["node_name", "port_name", "fabric_name"]
    for name in name_list:
        fc_file = os.path.join(fc_host_path, name)
        with open(fc_file, "r") as fc_content:
            fc_dict[name] = fc_content.read().strip().split("0x")[1]

    # Check wwnn, wwpn and fabric_wwn
    if len(wwnn_from_xml) != 16:
        logging.error("The wwnn is not valid: %s", wwnn_from_xml)
        return False
    if len(wwpn_from_xml) != 16:
        logging.error("The wwpn is not valid: %s", wwpn_from_xml)
        return False
    if fc_dict["node_name"] != wwnn_from_xml:
        logging.error("The node name is differnet: %s is not %s",
                      fc_dict["node_name"], wwnn_from_xml)
        return False
    if fc_dict["port_name"] != wwpn_from_xml:
        logging.error("The port name is different: %s is not %s",
                      fc_dict["port_name"], wwpn_from_xml)
        return False
    if fc_dict["fabric_name"] != fabric_wwn_from_xml:
        logging.error("The fabric wwpn is differnt: %s is not %s",
                      fc_dict["fabric_name"], fabric_wwn_from_xml)
        return False

    fc_type_from_xml = dev_xml.fc_type
    cap_type_from_xml = dev_xml.cap_type

    # Check capability type
    if (cap_type_from_xml != "scsi_host") or (fc_type_from_xml != "fc_host"):
        logging.error("The capability type isn't 'scsi_host' or 'fc_host'")
        return False

    return True


def find_hbas(hba_type="hba", status="online"):
    """
    Find online hba/vhba cards.

    :params hba_type: "vhba" or "hba"
    :params status: "online" or "offline"
    :return: A list contains the online/offline vhba/hba list
    """
    # TODO: add status=offline/online judgement, we don't test offline vhba now
    # so leave it here as a placeholder.
    result = virsh.nodedev_list(cap="scsi_host")
    if result.exit_status:
        raise exceptions.TestFail(result.stderr_text)
    scsi_hosts = result.stdout_text.strip().splitlines()
    online_hbas_list = []
    online_vhbas_list = []
    # go through all scsi hosts, and split hbas/vhbas into lists
    for scsi_host in scsi_hosts:
        result = virsh.nodedev_dumpxml(scsi_host)
        stdout = result.stdout_text.strip()
        if result.exit_status:
            raise exceptions.TestFail(result.stderr_text)
        if (re.search('vport_ops', stdout)
                and not re.search('<fabric_wwn>ffffffffffffffff</fabric_wwn>', stdout)
                and not re.search('<fabric_wwn>0</fabric_wwn>', stdout)):
            online_hbas_list.append(scsi_host)
        if re.search('fc_host', stdout) and not re.search('vport_ops', stdout):
            online_vhbas_list.append(scsi_host)
    if hba_type == "hba":
        return online_hbas_list
    if hba_type == "vhba":
        return online_vhbas_list


def is_vhbas_added(old_vhbas):
    """
    Check if a vhba is added

    :param old_vhbas: Pre-existing vhbas
    :return: True/False based on addition
    """
    new_vhbas = find_hbas("vhba")
    new_vhbas.sort()
    old_vhbas.sort()
    if len(new_vhbas) - len(old_vhbas) >= 1:
        return True
    return False


def is_vhbas_removed(old_vhbas):
    """
    Check if a vhba is removed

    :param old_vhbas: Pre-existing vhbas
    :return: True/False based on removal
    """
    new_vhbas = find_hbas("vhba")
    new_vhbas.sort()
    old_vhbas.sort()
    if len(new_vhbas) - len(old_vhbas) < 0:
        return True
    return False


def nodedev_create_from_xml(params):
    """
    Create a node device with a xml object.

    :param params: Including nodedev_parent, scsi_wwnn, scsi_wwpn set in xml
    :return: The scsi device name just created
    """
    nodedev_parent = params.get("nodedev_parent")
    scsi_wwnn = params.get("scsi_wwnn")
    scsi_wwpn = params.get("scsi_wwpn")
    status_error = "yes" == params.get("status_error", "no")
    vhba_xml = NodedevXML()
    vhba_xml.cap_type = 'scsi_host'
    vhba_xml.fc_type = 'fc_host'
    vhba_xml.parent = nodedev_parent
    vhba_xml.wwnn = scsi_wwnn
    vhba_xml.wwpn = scsi_wwpn
    logging.debug("Prepare the nodedev XML: %s", vhba_xml)
    vhba_file = mktemp()
    with open(vhba_file, 'w') as xml_object:
        xml_object.write(str(vhba_xml))

    result = virsh.nodedev_create(vhba_file,
                                  debug=True,
                                  )
    # Remove temprorary file
    os.unlink(vhba_file)
    libvirt.check_exit_status(result, status_error)
    output = result.stdout_text
    logging.info(output)
    for scsi in output.split():
        if scsi.startswith('scsi_host'):
            # Check node device
            utils_misc.wait_for(
                lambda: check_nodedev(scsi, nodedev_parent),
                timeout=_TIMEOUT)
            if check_nodedev(scsi, nodedev_parent):
                return scsi
            else:
                raise exceptions.TestFail(
                    "XML of vHBA card '%s' is not correct,"
                    "Please refer to log err for detailed info" % scsi)


def nodedev_destroy(scsi_host, params={}):
    """
    Destroy a nodedev of scsi_host#.
    :param scsi_host: The scsi to destroy
    :param params: Contain status_error
    """
    status_error = "yes" == params.get("status_error", "no")
    result = virsh.nodedev_destroy(scsi_host)
    logging.info("destroying scsi:%s", scsi_host)
    # Check status_error
    libvirt.check_exit_status(result, status_error)
    # Check nodedev value
    if not check_nodedev(scsi_host):
        logging.info(result.stdout_text)
    else:
        raise exceptions.TestFail("The relevant directory still exists"
                                  " or mismatch with result")


def check_nodedev_exist(scsi_host):
    """
    Check if scsi_host# exist.
    :param scsi_host: The scsi host to be checked
    :return: True if scsi_host exist, False if not
    """
    host = scsi_host.split("_")[1]
    fc_host_path = os.path.join(_FC_HOST_PATH, host)
    if os.path.exists(fc_host_path):
        return True
    return False


def vhbas_cleanup(vhba_list):
    """
    Clean up vhbas.
    """
    for scsi_host in vhba_list:
        nodedev_destroy(scsi_host)
    left_vhbas = find_hbas("vhba")
    if left_vhbas:
        logging.error("old vhbas are: %s", left_vhbas)
    else:
        logging.debug("scsi_hosts destroyed: %s", vhba_list)


def create_hostdev_xml(adapter_name="", **kwargs):
    """
    Create vhba hostdev xml.

    :param adapter_name: The name of the scsi adapter
    :param kwargs: Could contain addr_bus, addr_target,
     addr_unit, mode, and managed
    :return: a xml object set by kwargs
    """
    addr_bus = kwargs.get('addr_bus', 0)
    addr_target = kwargs.get('addr_target', 0)
    addr_unit = kwargs.get('addr_unit', 0)
    mode = kwargs.get('mode', 'subsystem')
    managed = kwargs.get('managed', 'no')

    hostdev_xml = hostdev.Hostdev()
    hostdev_xml.type = "scsi"
    hostdev_xml.managed = managed
    hostdev_xml.mode = mode

    source_args = {}
    source_args['adapter_name'] = adapter_name
    source_args['bus'] = addr_bus
    source_args['target'] = addr_target
    source_args['unit'] = addr_unit
    hostdev_xml.source = hostdev_xml.new_source(**source_args)
    logging.info(hostdev_xml)
    return hostdev_xml


def find_scsi_luns(scsi_host):
    """
    Find available luns of specified scsi_host.

    :param scsi_host: The scsi host name in format of "scsi_host#"
    :return: A dictionary contains all available fc luns
    """
    lun_dicts = []
    tmp_list = []
    scsi_number = scsi_host.replace("scsi_host", "")
    cmd = "multipath -ll | grep '\- %s:' | grep 'ready running' |\
           awk '{FS=\" \"}{for (f=1; f<=NF; f+=1) {if ($f ~ /%s:/)\
           {print $f}}}'" % (scsi_number, scsi_number)
    try:
        result = process.run(cmd, shell=True)
    except Exception as e:
        raise exceptions.TestError("run 'multipath' failed: %s" % str(e))
    tmp_list = result.stdout_text.strip().splitlines()
    for lun in tmp_list:
        lun = lun.split(":")
        lun_dicts_item = {}
        lun_dicts_item["scsi"] = lun[0]
        lun_dicts_item["bus"] = lun[1]
        lun_dicts_item["target"] = lun[2]
        lun_dicts_item["unit"] = lun[3]
        lun_dicts.append(lun_dicts_item)
    return lun_dicts


def find_mpath_devs():
    """
    Find all mpath devices in /dev/mapper which is start with "mpath"
    and not ending with a digit (which means it's a partition)
    """
    mpath_devs = []
    cmd = "ls -l /dev/mapper/ | grep mpath | awk -F ' ' '{print $9}' \
           | grep -Ev [0-9]$ |sort -d"
    cmd_result = process.run(cmd, shell=True)
    mpath_devs = cmd_result.stdout_text.split("\n")
    return mpath_devs


def is_mpath_devs_added(old_mpath_devs):
    """
    Check if a mpath device is added
    :param old_mpaths: Pre-existing mpaths
    :return: True/False based on addition
    """
    new_mpath_devs = find_mpath_devs()
    new_mpath_devs.sort()
    old_mpath_devs.sort()
    if len(new_mpath_devs) - len(old_mpath_devs) >= 1:
        return True
    return False


def restart_multipathd(mpath_dev="", expect_exist=False):
    """
    Restart the multipath deamon, and check if mpath_dev still exists
    after deamon restarted, as expected.
    """
    cmd_status = process.system('service multipathd restart', verbose=True)
    if cmd_status:
        raise exceptions.TestFail("Restart multipathd failed.")
    if not os.path.exists(mpath_dev):
        if not expect_exist:
            return True
    else:
        if expect_exist:
            return True
    return False


def prepare_multipath_conf(conf_path="/etc/multipath.conf", conf_content="",
                           replace_existing=False, restart_multipath=True):
    """
    Prepare multipath conf file.

    :param conf_path: Path to the conf file.
    :param conf_content: Content of the conf file.
    :param replace_existing: True means to replace exsiting conf file.
    :param restart_multipathd: True means to restart multipathd.
    :return: The content of original conf, can be used to recover env.
    """
    default_conf_content = ("defaults {\n\tuser_friendly_names yes"
                            "\n\tfind_multipaths yes\n}")
    old_conf_content = ""
    new_conf_content = conf_content if conf_content else default_conf_content
    if os.path.exists(conf_path):
        with open(conf_path, 'r+') as conf_file:
            old_conf_content = conf_file.read()
            logging.info("Old multipath conf is: %s" % old_conf_content)
            if replace_existing:
                conf_file.seek(0)
                conf_file.truncate()
                conf_file.write(new_conf_content)
                logging.info("Replace multipath conf to: %s" % new_conf_content)
            else:
                logging.info("Multipath conf exsits, skip preparation.")
    else:
        with open(conf_path, 'w') as conf_file:
            conf_file.write(new_conf_content)
            logging.info("Create multipath conf: %s" % new_conf_content)
    if restart_multipath:
        restart_multipathd()
    return old_conf_content



from __future__ import division
import re
import math
from decimal import Decimal
from decimal import getcontext


def align_value(value, factor=1024):
    """
    Value will be factored to the mentioned number.

    :param value: value to be checked or changed to factor
    :param factor: value used for align
    :return: aligned value
    """
    return int(math.ceil(float(value) / factor) * factor)


def format_size_human_readable(value, binary=False, precision='%.2f'):
    """
    Format a number of bytesize to a human readable filesize.

    By default,decimal suffixes and base:10**3 will be used.
    :param binary: use binary suffixes (KiB, MiB) and use 2*10 as base
    :param precision: format string to specify precision for float
    """
    suffixes = {
        'decimal': ('B', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB'),
        'binary': ('B', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB', 'YiB')
    }
    suffix = suffixes['binary'] if binary else suffixes['decimal']
    base = 1024 if binary else 1000

    value = float(value)
    for i, s in enumerate(suffix):
        unit = base ** (i + 1)
        if value < unit:
            break
    value = value * base / unit
    format_str = ('%d' if value.is_integer() else precision) + ' %s'
    return format_str % (value, s)


def normalize_data_size(value_str, order_magnitude="M", factor=1024):
    """
    Normalize a data size in one order of magnitude to another.

    :param value_str: a string include the data default unit is 'B'
    :param order_magnitude: the magnitude order of result
    :param factor: int, the factor between two relative order of magnitude.
                   Normally could be 1024 or 1000
    :return normalized data size string
    """
    def _get_unit_index(m):
        try:
            return ['B', 'K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y'].index(
                m.upper())
        except ValueError:
            pass
        return 0

    def _trim_tailling_zeros(num_str):
        # remove tailing zeros, convert float number str to int str
        if '.' in num_str:
            num_str = num_str.rstrip('0').rstrip('.')
        return num_str

    regex = r"(\d+\.?\d*)\s*(\w?)"
    match = re.search(regex, value_str)
    try:
        value = match.group(1)
        unit = match.group(2)
        if not unit:
            unit = 'B'
    except TypeError:
        raise ValueError("Invalid data size format 'value_str=%s'" % value_str)

    getcontext().prec = 20
    from_index = _get_unit_index(unit)
    to_index = _get_unit_index(order_magnitude)
    if from_index - to_index >= 0:
        d = Decimal(value) * Decimal(factor ** (from_index - to_index))
    else:
        d = Decimal(value) / Decimal(factor ** (to_index - from_index))
    return _trim_tailling_zeros('{:f}'.format(d))



"""
This script is used to execute a program and collect the monitor
information in background, redirect the outputs to log files.
"""

import threading
import shelve
import re
import os
import sys
import random
import string


def getstatusoutput(cmd):
    """Return (status, output) of executing cmd in a shell."""
    pipe = os.popen('{ ' + cmd + '; } 2>&1', 'r')
    text = pipe.read()
    sts = pipe.close()
    if sts is None:
        sts = 0
    if text[-1:] == '\n':
        text = text[:-1]
    return sts, text


class Runner(object):

    def __init__(self):
        """
        Set the global parameter for thread clean up
        """
        self.kill_thread_flag = False

    def monitor_thread(self, m_cmd, p_file, r_path):
        """
        Record the parent process id and start the monitor process
        in background
        """
        fd = shelve.open(p_file)
        fd["pid"] = os.getpid()
        fd.close()
        getstatusoutput("%s &> %s_monitor" % (m_cmd, r_path))

    def thread_kill(self, cmd, p_file):
        """
        Kill the process according to its parent pid and command
        """
        fd = shelve.open(p_file)
        _, o = getstatusoutput("pstree -p %s" % fd["pid"])
        try:
            tmp = cmd.split()[0]
            pid = re.findall("%s.(\d+)" % tmp, o)[0]
        except IndexError:
            return (0, "")
        s, o = getstatusoutput("kill -9 %s" % pid)
        fd.close()
        return (s, o)

    def test_thread(self, m_cmd, t_cmd, p_file):
        """
        Test thread
        """
        self.kill_thread_flag = True
        s, o = getstatusoutput(t_cmd)
        if s != 0:
            print("Test failed or timeout: %s" % o)
        if self.kill_thread_flag:
            s, o = self.thread_kill(m_cmd, p_file)
            if s != 0:
                print("Monitor process is still alive, %s" % o)
            else:
                self.kill_thread_flag = False

    def run(self, m_cmd, t_cmd, r_path, timeout):
        """
        Main thread for testing, will do clean up afterwards
        """
        pid_file = "/tmp/pid_file_%s" % "".join(random.sample(string.letters,
                                                              4))
        monitor = threading.Thread(target=self.monitor_thread, args=(m_cmd,
                                                                     pid_file, r_path))
        test_runner = threading.Thread(target=self.test_thread, args=(m_cmd,
                                                                      t_cmd, pid_file))

        monitor.start()
        test_runner.start()
        monitor.join(timeout)
        if self.kill_thread_flag:
            self.thread_kill(m_cmd, pid_file)
            self.thread_kill(t_cmd, pid_file)
            self.kill_thread_flag = False


if __name__ == '__main__':
    if len(sys.argv) < 4:
        this = os.path.basename(sys.argv[0])
        print("Usage: %s <monitor_cmd> <test_cmd> <test_path> <timeout>" % this)
        sys.exit(1)

    monitor_cmd = sys.argv[1]
    test_cmd = sys.argv[2]
    test_path = sys.argv[3]
    test_cmd = test_cmd % test_path
    timeout = int(sys.argv[4])
    r = Runner()
    r.run(monitor_cmd, test_cmd, test_path, timeout)
